@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/configedit.c 1.17 2002/04/07 12:31:46 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/configedit.c 1.31 2003/08/25 13:34:55 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7b.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   The HTML interactive configuration editing pages.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -22,120 +22,19 @@
 #include <unistd.h>
 
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
 #include "configpriv.h"
 #include "config.h"
-#include "errors.h"
 
 
 static void ConfigurationIndexPage(int fd,/*@null@*/ char *url);
 static void ConfigurationSectionPage(int fd,int section,/*@null@*/ char *url);
-static void ConfigurationItemPage(int fd,int section,int item,/*@null@*/ char *url,Body *request_body);
-static void ConfigurationURLPage(int fd,/*@null@*/ char *request_args);
+static void ConfigurationItemPage(int fd,int section,int item,/*@null@*/ char *url,/*@null@*/ Body *request_body);
+static void ConfigurationEditURLPage(int fd,/*@null@*/ Body *request_body);
+static void ConfigurationURLPage(int fd,/*@null@*/ char *url);
 static void ConfigurationAuthFail(int fd,char *url);
 
-/*+ The list of configuration items to list on the configuration page for a URL. +*/
-static ConfigItem *URLConfigItems[]={
-
- /* OnlineOptions */
-
- &RequestChanged,
- &RequestChangedOnce,
- &RequestExpired,
- &RequestNoCache,
- &RequestRedirection,
- &TryWithoutPassword,
- &IntrDownloadKeep,
- &IntrDownloadSize,
- &IntrDownloadPercent,
- &TimeoutDownloadKeep,
- &RequestCompressedData,
-
- /* OfflineOptions */
-
- &PragmaNoCache,
- &ConfirmRequests,
- &DontRequestOffline,
-
- /* FetchOptions */
-
- &FetchStyleSheets,
- &FetchImages,
- &FetchWebbugImages,
- &FetchFrames,
- &FetchScripts,
- &FetchObjects,
-
- /* IndexOptions */
-
- &IndexListOutgoing,
- &IndexListLatest,
- &IndexListMonitor,
- &IndexListHost,
- &IndexListAny,
-
- /* ModifyHtml */
-
- &EnableHTMLModifications,
- &EnableModificationsOnline,
- &AddCacheInfo,
- &AnchorModifyBegin[0],
- &AnchorModifyBegin[1],
- &AnchorModifyBegin[2],
- &AnchorModifyEnd[0],
- &AnchorModifyEnd[1],
- &AnchorModifyEnd[2],
- &DisableHTMLScript,
- &DisableHTMLApplet,
- &DisableHTMLStyle,
- &DisableHTMLBlink,
- &DisableHTMLFlash,
- &DisableHTMLMetaRefresh,
- &DisableHTMLMetaRefreshSelf,
- &DisableHTMLDontGetAnchors,
- &DisableHTMLDontGetIFrames,
- &ReplaceHTMLDontGetImages,
- &ReplacementHTMLDontGetImage,
- &ReplaceHTMLWebbugImages,
- &ReplacementHTMLWebbugImage,
- &DemoroniseMSChars,
- &DisableAnimatedGIF,
-
- /* DontCache */
-
- &DontCache,
-
- /* DontGet */
-
- &DontGet,
- &DontGetReplacementURL,
- &DontGetRecursive,
- &DontGetLocation,
-
- /* CensorHeader */
-
- &CensorHeader,
- &RefererSelf,
- &RefererSelfDir,
-
- /* FTPOptions */
-
- &FTPAuthUser,
- &FTPAuthPass,
-
- /* Proxy */
-
- &Proxies,
- &ProxyAuthUser,
- &ProxyAuthPass,
- &SSLProxy,
-
- /* Purge */
-
- &PurgeAges,
- &PurgeCompressAges
-};
-
 
 /*++++++++++++++++++++++++++++++++++++++
   Send to the client one of the pages to configure WWWOFFLE using HTML.
@@ -168,30 +67,10 @@ void ConfigurationPage(int fd,URL *Url,Body *request_body)
       }
    }
 
- /* Parse the arguments if any */
+ /* Extract the URL argument if any (ignoring all posted forms) */
 
- if(Url->args)
-   {
-    char **args,*pling;
-
-    if(*Url->args=='!' && (pling=strstr(Url->args,"!POST")))
-      {
-       *pling=0;
-       args=SplitFormArgs(Url->args+1);
-       *pling='!';
-      }
-    else
-       args=SplitFormArgs(Url->args);
-
-    for(i=0;args[i];i++)
-      {
-       if(!strncmp("url=",args[i],4) && args[i][4])
-          url=URLDecodeFormArgs(args[i]+4);
-      }
-
-    free(args[0]);
-    free(args);
-   }
+ if(Url->args && *Url->args!='!')
+    url=URLDecodeFormArgs(Url->args);
 
  /* Determine the page to show. */
 
@@ -202,9 +81,16 @@ void ConfigurationPage(int fd,URL *Url,Body *request_body)
     return;
    }
 
+ if(!strcmp(&Url->path[15],"editurl"))
+   {
+    ConfigurationEditURLPage(fd,request_body);
+    if(url) free(url);
+    return;
+   }
+
  if(!strcmp(&Url->path[15],"url"))
    {
-    ConfigurationURLPage(fd,Url->args);
+    ConfigurationURLPage(fd,url);
     if(url) free(url);
     return;
    }
@@ -252,18 +138,20 @@ static void ConfigurationIndexPage(int fd,char *url)
 {
  char *line=NULL;
  int file;
+ off_t seekpos=0;
  int s;
 
  file=OpenLanguageFile("messages/README.CONF.txt");
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
     return;
    }
 
+ init_io(file);
 
  line=read_line(file,line); /* TITLE ... */
  line=read_line(file,line); /* HEAD */
@@ -280,7 +168,7 @@ static void ConfigurationIndexPage(int fd,char *url)
     char *description=NULL;
 
     lseek(file,0,SEEK_SET);
-    init_buffer(file);
+    reinit_io(file);
 
     while((line=read_line(file,line)))
       {
@@ -292,6 +180,11 @@ static void ConfigurationIndexPage(int fd,char *url)
           if(line)
              line[strlen(line)-1]=0;
           description=line;
+          seekpos=lseek(file,0,SEEK_CUR);
+          if(seekpos>READ_BUFFER_SIZE)
+             seekpos-=READ_BUFFER_SIZE;
+          else
+             seekpos=0;
           break;
          }
       }
@@ -303,7 +196,8 @@ static void ConfigurationIndexPage(int fd,char *url)
                     NULL);
    }
 
- lseek(file,0,SEEK_SET);
+ lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the last section */
+ reinit_io(file);
 
  while((line=read_line(file,line)))
    {
@@ -316,6 +210,12 @@ static void ConfigurationIndexPage(int fd,char *url)
  HTMLMessageBody(fd,"ConfigurationPage-Tail",
                  "description",line,
                  NULL);
+
+ if(line)
+    free(line);
+
+ finish_io(file);
+ close(file);
 }
 
 
@@ -340,12 +240,13 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
     return;
    }
 
+ init_io(file);
 
  HTMLMessageHead(fd,200,"WWWOFFLE Configuration Section Page",
                  NULL);
@@ -386,8 +287,8 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
     char *template=NULL;
     description=NULL;
 
-    lseek(file,seekpos,SEEK_SET);
-    init_buffer(file);
+    lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the required section */
+    reinit_io(file);
 
     while((line1=read_line(file,line1)))
       {
@@ -401,6 +302,8 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
       {
        line1[strlen(line1)-1]=0;
 
+       if(!strncmp(line1,"SECTION",7))
+          break;
        if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[section]->itemdefs[i].name))
          {
           line1=read_line(file,line1);
@@ -429,6 +332,9 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
     free(line1);
  if(line2)
     free(line2);
+
+ finish_io(file);
+ close(file);
 }
 
 
@@ -448,7 +354,7 @@ static void ConfigurationSectionPage(int fd,int section,char *url)
 
 static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *request_body)
 {
- char *edit=NULL,*action=NULL,*entry=NULL;
+ char *action=NULL,*entry=NULL;
  char *key=NULL,*val=NULL;
  int i;
 
@@ -462,15 +368,15 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
     for(i=0;args[i];i++)
       {
        if(!strncmp("url=",args[i],4) && args[i][4])
-          url=URLDecodeFormArgs(args[i]+4);
+          url=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("key=",args[i],4) && args[i][4])
-          key=URLDecodeFormArgs(args[i]+4);
+          key=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("val=",args[i],4) && args[i][4])
-          val=URLDecodeFormArgs(args[i]+4);
+          val=TrimArgs(URLDecodeFormArgs(args[i]+4));
        if(!strncmp("action=",args[i],7))
-          action=URLDecodeFormArgs(args[i]+7);
+          action=TrimArgs(URLDecodeFormArgs(args[i]+7));
        if(!strncmp("entry=",args[i],6))
-          entry=URLDecodeFormArgs(args[i]+6);
+          entry=TrimArgs(URLDecodeFormArgs(args[i]+6));
       }
 
     free(args[0]);
@@ -486,9 +392,12 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
     int file;
     char nentries[8],nallowed[8];
     char *entry_url=NULL,*entry_key=NULL,*entry_val=NULL;
+    int edit=0;
 
     if(action && entry)
       {
+       edit=1;
+
        for(i=0;i<(*CurrentConfig.sections[section]->itemdefs[item].item)->nentries;i++)
          {
           char *thisentry=ConfigEntryString(*CurrentConfig.sections[section]->itemdefs[item].item,i);
@@ -510,12 +419,15 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
 
     if(file==-1)
       {
-       HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+       HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                    "error","Cannot open README.CONF.txt",
                    NULL);
+
+       if(entry) free(entry);
        return;
       }
 
+    init_io(file);
 
     HTMLMessageHead(fd,200,"WWWOFFLE Configuration Item Page",
                     NULL);
@@ -639,6 +551,9 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
        free(line1);
     if(line2)
        free(line2);
+
+    finish_io(file);
+    close(file);
    }
 
     /* Display the page with the results of editing the parameters. */
@@ -685,66 +600,58 @@ static void ConfigurationItemPage(int fd,int section,int item,char *url,Body *re
 
  if(key)    free(key);
  if(val)    free(val);
- if(edit)   free(edit);
  if(action) free(action);
  if(entry)  free(entry);
 }
 
 
 /*++++++++++++++++++++++++++++++++++++++
-  The configuration page for a specific URL.
+  The configuration page to edit the URL-SPECIFICATION and redirect to the real page.
 
   int fd The file descriptor to write to.
 
-  char *request_args The arguments to the configuration page.
+  Body *request_body The body of the request.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void ConfigurationURLPage(int fd,char *request_args)
+static void ConfigurationEditURLPage(int fd,Body *request_body)
 {
- char *url=NULL;
+ char *url=NULL,*encurl;
+ int urllen;
  char *proto=NULL,*host=NULL,*port=NULL,*path=NULL,*args=NULL;
  char *proto_other=NULL,*host_other=NULL,*port_other=NULL,*path_other=NULL,*args_other=NULL;
- int wildcard=0;
- int file;
- char *line1=NULL,*line2=NULL;
- int configitem,s,last_s=-1,seekpos=0;
+ char *localhost,*relocate;
 
  /* Parse the arguments. */
 
- if(request_args)
+ if(request_body)
    {
-    char *form,**arglist;
+    char *form=URLRecodeFormArgs(request_body->content);
+    char **arglist=SplitFormArgs(form);
     int i;
 
-    form=URLRecodeFormArgs(request_args);
-    arglist=SplitFormArgs(form);
-
     for(i=0;arglist[i];i++)
       {
-       if(!strncmp("url=",arglist[i],4) && arglist[i][4])
-          url=URLDecodeFormArgs(arglist[i]+4);
-
        if(!strncmp("proto=",arglist[i],6) && arglist[i][6])
-          proto=URLDecodeFormArgs(arglist[i]+6);
+          proto=TrimArgs(URLDecodeFormArgs(arglist[i]+6));
        if(!strncmp("host=",arglist[i],5) && arglist[i][5])
-          host=URLDecodeFormArgs(arglist[i]+5);
+          host=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
        if(!strncmp("port=",arglist[i],5) && arglist[i][5])
-          port=URLDecodeFormArgs(arglist[i]+5);
+          port=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
        if(!strncmp("path=",arglist[i],5) && arglist[i][5])
-          path=URLDecodeFormArgs(arglist[i]+5);
+          path=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
        if(!strncmp("args=",arglist[i],5) && arglist[i][5])
-          args=URLDecodeFormArgs(arglist[i]+5);
+          args=TrimArgs(URLDecodeFormArgs(arglist[i]+5));
 
        if(!strncmp("proto_other=",arglist[i],12) && arglist[i][12])
-          proto_other=URLDecodeFormArgs(arglist[i]+12);
+          proto_other=TrimArgs(URLDecodeFormArgs(arglist[i]+12));
        if(!strncmp("host_other=",arglist[i],11) && arglist[i][11])
-          host_other=URLDecodeFormArgs(arglist[i]+11);
+          host_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
        if(!strncmp("port_other=",arglist[i],11) && arglist[i][11])
-          port_other=URLDecodeFormArgs(arglist[i]+11);
+          port_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
        if(!strncmp("path_other=",arglist[i],11) && arglist[i][11])
-          path_other=URLDecodeFormArgs(arglist[i]+11);
+          path_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
        if(!strncmp("args_other=",arglist[i],11) && arglist[i][11])
-          args_other=URLDecodeFormArgs(arglist[i]+11);
+          args_other=TrimArgs(URLDecodeFormArgs(arglist[i]+11));
       }
 
     free(arglist[0]);
@@ -754,104 +661,148 @@ static void ConfigurationURLPage(int fd,char *request_args)
 
  /* Sort out a URL from the mess of arguments. */
 
- if(url)
-   {
-    char *colon;
-    URL *Url=SplitURL(url);
-
-    if(proto_other)free(proto_other);
-    if(host_other) free(host_other);
-    if(port_other) free(port_other);
-    if(path_other) free(path_other);
-    if(args_other) free(args_other);
-
-    proto=(char*)realloc((void*)proto,strlen(Url->proto)+1);
-    strcpy(proto,Url->proto);
-
-    colon=strchr(Url->host,':');
-
-    host=(char*)realloc((void*)host,strlen(Url->host)+1);
-    strcpy(host,Url->host);
-    if(colon)
-       host[colon-Url->host]=0;
-
-    if(colon)
-      {
-       port=(char*)realloc((void*)port,strlen(colon+1)+1);
-       strcpy(port,colon+1);
-      }
-
-    path=(char*)realloc((void*)path,strlen(Url->path)+1);
-    strcpy(path,Url->path);
-
-    if(Url->args)
-      {
-       args=(char*)realloc((void*)args,strlen(Url->args)+1);
-       strcpy(args,Url->args);
-      }
-
-    FreeURL(Url);
-   }
+ if(proto && !strcmp(proto,"OTHER") && proto_other)
+   {free(proto); proto=proto_other;}
  else
+   {if(proto_other) free(proto_other);}
+
+ if(host && !strcmp(host,"OTHER") && host_other)
+   {free(host); host=host_other;}
+ else
+   {if(host_other) free(host_other);}
+
+ if(port && !strcmp(port,"OTHER") && port_other)
+   {free(port); port=(char*)malloc(strlen(port_other)+2); strcpy(port,":"); strcat(port,port_other);}
+ else
+   {if(port_other) free(port_other);}
+
+ if(path && !strcmp(path,"OTHER") && path_other)
+   {free(path); path=path_other;}
+ else
+   {if(path_other) free(path_other);}
+
+ if(args && !strcmp(args,"OTHER") && args_other)
+   {free(args); args=(char*)malloc(strlen(args_other)+2); strcpy(args,"?"); strcat(args,args_other);}
+ else
+   {if(args_other) free(args_other);}
+
+ if(!proto)
+    proto=strcpy(malloc(2),"*");
+
+ if(!host)
+    host=strcpy(malloc(2),"*");
+
+ if(!path)
+    path=strcpy(malloc(2),"*");
+
+ /* Create the URL */
+
+ urllen=4;
+ urllen+=strlen(proto);
+ urllen+=strlen(host);
+ if(port)
+    urllen+=strlen(port);
+ urllen+=strlen(path);
+ if(args)
+    urllen+=strlen(args);
+
+ url=(char*)malloc(urllen);
+
+ strcpy(url,proto);
+ strcat(url,"://");
+ strcat(url,host);
+ if(port)
+    strcat(url,port);
+ strcat(url,path);
+ if(args)
+    strcat(url,args);
+
+ if(proto)free(proto);
+ if(host) free(host);
+ if(port) free(port);
+ if(path) free(path);
+ if(args) free(args);
+
+ /* Redirect the browser to it */
+
+ localhost=GetLocalHost(1);
+ encurl=URLEncodeFormArgs(url);
+ relocate=(char*)malloc(strlen(localhost)+strlen(encurl)+32);
+
+ sprintf(relocate,"http://%s/configuration/url?%s",localhost,encurl);
+
+ HTMLMessage(fd,302,"WWWOFFLE Configuration Edit URL Redirect",relocate,"Redirect",
+             "location",relocate,
+             NULL);
+
+ free(url);
+ free(encurl);
+ free(localhost);
+ free(relocate);
+}
+
+
+/*++++++++++++++++++++++++++++++++++++++
+  The configuration page for a specific URL.
+
+  int fd The file descriptor to write to.
+
+  char *url The URL specification from the URL argument.
+  ++++++++++++++++++++++++++++++++++++++*/
+
+static void ConfigurationURLPage(int fd,char *url)
+{
+ char *proto=NULL,*host=NULL,*port=NULL,*path=NULL,*args=NULL;
+ char *copy,*colon,*slash,*ques;
+ int wildcard=0,file,s,i,seekpos=0;
+ char *line1=NULL,*line2=NULL;
+
+ /* Assume a "well-formed" URL, from the function above or a WWWOFFLE index. */
+
+ copy=(char*)malloc(strlen(url)+1);
+ strcpy(copy,url);
+
+ colon=strchr(copy,':');
+ slash=strchr(copy,'/');
+
+ proto=copy;
+
+ host=slash+2;
+
+ if(colon && slash && slash==colon+1 && *(slash+1)=='/')
    {
-    int urllen;
+    *colon=0;
 
-    if(proto && !strcmp(proto,"OTHER") && proto_other)
-      {free(proto); proto=proto_other;}
-    else
-      {if(proto_other) free(proto_other);}
-
-    if(host && !strcmp(host,"OTHER") && host_other)
-      {free(host); host=host_other;}
-    else
-      {if(host_other) free(host_other);}
-
-    if(port && !strcmp(port,"OTHER") && port_other)
-      {free(port); port=(char*)malloc(strlen(port_other)+1); strcpy(port,":"); strcat(port,port_other);}
-    else
-      {if(port_other) free(port_other);}
-
-    if(path && !strcmp(path,"OTHER") && path_other)
-      {free(path); path=path_other;}
-    else
-      {if(path_other) free(path_other);}
-
-    if(args && !strcmp(args,"OTHER") && args_other)
-      {free(args); args=(char*)malloc(strlen(args_other)+1); strcpy(args,"?"); strcat(args,args_other);}
-    else
-      {if(args_other) free(args_other);}
-
-    if(!proto)
-       proto=strcpy(malloc(2),"*");
-
-    if(!host)
-       host=strcpy(malloc(2),"*");
-
-    if(!path)
-       path=strcpy(malloc(2),"*");
-
-    urllen=4;
-    urllen+=strlen(proto);
-    urllen+=strlen(host);
-    if(port)
-       urllen+=strlen(port);
-    urllen+=strlen(path);
-    if(args)
-       urllen+=strlen(args);
-
-    url=(char*)realloc((void*)url,urllen);
-
-    strcpy(url,proto);
-    strcat(url,"://");
-    strcat(url,host);
-    if(port)
-       strcat(url,port);
-    strcat(url,path);
-    if(args)
-       strcat(url,args);
+    colon=strchr(slash+2,':');
+    slash=strchr(slash+2,'/');
    }
 
- if(strchr(url,'*'))
+ if(colon && slash && colon<slash)
+   {
+    *colon=0;
+    if(slash==colon+1)
+       port=":";
+    else
+       port=colon+1;
+   }
+
+ *slash=0;
+
+ path=slash+1;
+
+ ques=strchr(path,'?');
+
+ if(ques)
+   {
+    if(*(ques+1)==0)
+       args="?";
+    else
+       args=ques+1;
+
+    *ques=0;
+   }
+
+ if(strchr(url,'*') || (port && *port==':') || (args && *args=='?'))
     wildcard=1;
 
  /* Display the HTML */
@@ -860,20 +811,17 @@ static void ConfigurationURLPage(int fd,char *request_args)
 
  if(file==-1)
    {
-    HTMLMessage(fd,200,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
+    HTMLMessage(fd,500,"WWWOFFLE Configuration Page Error",NULL,"ServerError",
                 "error","Cannot open README.CONF.txt",
                 NULL);
 
-    if(proto)free(proto);
-    if(host) free(host);
-    if(port) free(port);
-    if(path) free(path);
-    if(args) free(args);
-    if(url)  free(url);
+    free(copy);
 
     return;
    }
 
+ init_io(file);
+
  HTMLMessageHead(fd,200,"WWWOFFLE Configuration URL Page",
                  NULL);
 
@@ -882,144 +830,125 @@ static void ConfigurationURLPage(int fd,char *request_args)
                  "url",url,
                  "proto",proto,
                  "host",host,
-                 "port",port&&*port==':'?port+1:port,
+                 "port",port,
                  "path",path,
-                 "args",args&&*args=='?'?args+1:args,
+                 "args",args,
                  NULL);
 
- /* Loop through all of the specified ConfigItems. */
+ free(copy);
 
- for(configitem=0;configitem<sizeof(URLConfigItems)/sizeof(URLConfigItems[0]);configitem++)
+ /* Loop through all of the ConfigItems and find those that take a URL argument. */
+
+ for(s=0;s<CurrentConfig.nsections;s++)
    {
-    int i;
-    char *template=NULL,*description=NULL,*current=NULL;
-
-    for(s=0,i=0;s<CurrentConfig.nsections;s++)
-      {
-       for(i=0;i<CurrentConfig.sections[s]->nitemdefs;i++)
-          if(URLConfigItems[configitem]==CurrentConfig.sections[s]->itemdefs[i].item)
-             break;
-
-       if(i!=CurrentConfig.sections[s]->nitemdefs)
-          break;
-      }
-
-    if(s==CurrentConfig.nsections)
-       continue;
-
-    if(s!=last_s)
-      {
-       lseek(file,0,SEEK_SET);
-
-       while((line1=read_line(file,line1)))
-         {
-          line1[strlen(line1)-1]=0;
-
-          if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
-            {
-             line1=read_line(file,line1);
-             if(line1)
-                line1[strlen(line1)-1]=0;
-             description=line1;
-             seekpos=lseek(file,0,SEEK_CUR);
-             if(seekpos>READ_BUFFER_SIZE)
-                seekpos-=READ_BUFFER_SIZE;
-             else
-                seekpos=0;
-             break;
-            }
-         }
-
-       last_s=s;
-      }
-    else
-      {
-       lseek(file,seekpos,SEEK_SET);
-       init_buffer(file);
-
-       while((line1=read_line(file,line1)))
-         {
-          line1[strlen(line1)-1]=0;
-
-          if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
-             break;
-         }
-      }
+    lseek(file,0,SEEK_SET);  /* go back to the start of the file */
+    reinit_io(file);
 
     while((line1=read_line(file,line1)))
       {
        line1[strlen(line1)-1]=0;
 
-       if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[s]->itemdefs[i].name))
+       if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
          {
-          line1=read_line(file,line1);
-          if(line1)
-             line1[strlen(line1)-1]=0;
-          line2=read_line(file,line2);
-          if(line2)
-             line2[strlen(line2)-1]=0;
-          template=line1;
-          description=line2;
+          seekpos=lseek(file,0,SEEK_CUR);
+          if(seekpos>READ_BUFFER_SIZE)
+             seekpos-=READ_BUFFER_SIZE;
+          else
+             seekpos=0;
           break;
          }
       }
 
-    if(!wildcard)
-      {
-       URL *Url=SplitURL(url);
-
-       if(*URLConfigItems[configitem] &&
-          (*URLConfigItems[configitem])->itemdef->key_type!=String)
+    for(i=0;i<CurrentConfig.sections[s]->nitemdefs;i++)
+       if(CurrentConfig.sections[s]->itemdefs[i].url_type ||
+          CurrentConfig.sections[s]->itemdefs[i].key_type==UrlSpecification)
          {
-          int j;
+          char *template=NULL,*description=NULL,*current=NULL;
 
-          for(j=0;j<(*URLConfigItems[configitem])->nentries;j++)
-             if((*URLConfigItems[configitem])->itemdef->url_type)
+          lseek(file,seekpos,SEEK_SET);  /* go back only to the start of the current section */
+          reinit_io(file);
+
+          while((line1=read_line(file,line1)))
+            {
+             line1[strlen(line1)-1]=0;
+
+             if(!strncmp(line1,"SECTION",7) && !strcmp(line1+8,CurrentConfig.sections[s]->name))
+                break;
+            }
+
+          while((line1=read_line(file,line1)))
+            {
+             line1[strlen(line1)-1]=0;
+
+             if(!strncmp(line1,"SECTION",7))
+                break;
+             if(!strncmp(line1,"ITEM",4) && !strcmp(line1+5,CurrentConfig.sections[s]->itemdefs[i].name))
                {
-                if(!(*URLConfigItems[configitem])->url[j])
-                  break;
-                else if(MatchUrlSpecification((*URLConfigItems[configitem])->url[j],Url->proto,Url->host,Url->path,Url->args))
-                   break;
+                line1=read_line(file,line1);
+                if(line1)
+                   line1[strlen(line1)-1]=0;
+                line2=read_line(file,line2);
+                if(line2)
+                   line2[strlen(line2)-1]=0;
+                template=line1;
+                description=line2;
+                break;
                }
-             else
+            }
+
+          if(!wildcard)
+            {
+             URL *Url=SplitURL(url);
+
+             if(*CurrentConfig.sections[s]->itemdefs[i].item &&
+                CurrentConfig.sections[s]->itemdefs[i].key_type!=String)
                {
-                if(MatchUrlSpecification((*URLConfigItems[configitem])->key[j].urlspec,Url->proto,Url->host,Url->path,Url->args))
-                   break;
+                int j;
+
+                for(j=0;j<(*CurrentConfig.sections[s]->itemdefs[i].item)->nentries;j++)
+                   if(CurrentConfig.sections[s]->itemdefs[i].url_type)
+                     {
+                      if(!(*CurrentConfig.sections[s]->itemdefs[i].item)->url[j])
+                         break;
+                      else if(MatchUrlSpecification((*CurrentConfig.sections[s]->itemdefs[i].item)->url[j],Url->proto,Url->host,Url->path,Url->args))
+                         break;
+                     }
+                   else
+                     {
+                      if(MatchUrlSpecification((*CurrentConfig.sections[s]->itemdefs[i].item)->key[j].urlspec,Url->proto,Url->host,Url->path,Url->args))
+                         break;
+                     }
+
+                if(j!=(*CurrentConfig.sections[s]->itemdefs[i].item)->nentries)
+                   current=ConfigEntryString(*CurrentConfig.sections[s]->itemdefs[i].item,j);
                }
 
-          if(j!=(*URLConfigItems[configitem])->nentries)
-             current=ConfigEntryString(*URLConfigItems[configitem],j);
+             FreeURL(Url);
+            }
+
+          HTMLMessageBody(fd,"ConfigurationUrl-Body",
+                          "section",CurrentConfig.sections[s]->name,
+                          "item",CurrentConfig.sections[s]->itemdefs[i].name,
+                          "template",template,
+                          "description",description,
+                          "current",current,
+                          NULL);
+
+          if(current)
+             free(current);
          }
-
-       FreeURL(Url);
-      }
-
-    HTMLMessageBody(fd,"ConfigurationUrl-Body",
-                    "section",CurrentConfig.sections[s]->name,
-                    "item",CurrentConfig.sections[s]->itemdefs[i].name,
-                    "template",template,
-                    "description",description,
-                    "current",current,
-                    NULL);
-
-    if(current)
-       free(current);
    }
 
  HTMLMessageBody(fd,"ConfigurationUrl-Tail",
                  NULL);
 
- if(proto)free(proto);
- if(host) free(host);
- if(port) free(port);
- if(path) free(path);
- if(args) free(args);
- if(url)  free(url);
-
  if(line1)
     free(line1);
  if(line2)
     free(line2);
+
+ finish_io(file);
+ close(file);
 }
 
 
