@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/configmisc.c 1.11 2002/04/21 10:41:26 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/configmisc.c 1.19 2003/02/14 19:25:58 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7b.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   Configuration file data management functions.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -20,24 +20,19 @@
 #include <string.h>
 #include <ctype.h>
 
-#include <limits.h>
-#include <unistd.h>
-#include <sys/types.h>
 #include <pwd.h>
 #include <grp.h>
 
-#include "configpriv.h"
-#include "errors.h"
-#include "wwwoffle.h"
 #include "misc.h"
+#include "errors.h"
+#include "configpriv.h"
 
 
 /* Local functions */
 
-static void FreeKeyOrValue(KeyOrValue *keyval,ConfigType type);
-
-static char* sprintf_key_or_value(ConfigType type,KeyOrValue key_or_val);
+static /*@null@*/ char* sprintf_key_or_value(ConfigType type,KeyOrValue key_or_val);
 static /*@null@*/ char* sprintf_url_spec(UrlSpec *urlspec);
+static /*@null@*/ char *strstrn(const char *phaystack, const char *pneedle, size_t n, int nocase);
 
 
 /*+ The backup version of the config file. +*/
@@ -156,34 +151,22 @@ void RestoreBackupConfigFile(void)
 
 void PurgeBackupConfigFile(int restore_startup)
 {
- int s=0,i;
+ int s,i;
 
- if(restore_startup)
-   {
-    /* Restore the StartUp section. */
+ /* Purge all of the sections and restore StartUp if needed. */
 
-    for(i=0;i<CurrentConfig.sections[s]->nitemdefs;i++)
-      {
-       FreeConfigItem(*CurrentConfig.sections[s]->itemdefs[i].item);
-
-       *CurrentConfig.sections[s]->itemdefs[i].item=*BackupConfig.sections[s]->itemdefs[i].item;
-
-       free(BackupConfig.sections[s]->itemdefs[i].item);
-      }
-
-    free(BackupConfig.sections[s]->itemdefs);
-    free(BackupConfig.sections[s]);
-
-    s++;
-   }
-
- /* Purge all of the sections except StartUp. */
-
- for(s=1;s<BackupConfig.nsections;s++)
+ for(s=0;s<BackupConfig.nsections;s++)
    {
     for(i=0;i<BackupConfig.sections[s]->nitemdefs;i++)
       {
-       FreeConfigItem(*BackupConfig.sections[s]->itemdefs[i].item);
+       if(s==0 && restore_startup)
+         {
+          FreeConfigItem(*CurrentConfig.sections[s]->itemdefs[i].item);
+
+          *CurrentConfig.sections[s]->itemdefs[i].item=*BackupConfig.sections[s]->itemdefs[i].item;
+         }
+       else
+          FreeConfigItem(*BackupConfig.sections[s]->itemdefs[i].item);
 
        free(BackupConfig.sections[s]->itemdefs[i].item);
       }
@@ -267,7 +250,7 @@ void FreeConfigItem(ConfigItem item)
   ConfigType type The type of key or value.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void FreeKeyOrValue(KeyOrValue *keyval,ConfigType type)
+void FreeKeyOrValue(KeyOrValue *keyval,ConfigType type)
 {
  switch(type)
    {
@@ -341,10 +324,10 @@ int MatchUrlSpecification(UrlSpec *spec,char *proto,char *host,char *path,char *
  SplitHostPort(host,&hoststr,&portstr);
 
  if((!spec->proto || !strcmp(UrlSpecProto(spec),proto)) &&
-    (!spec->host || WildcardMatch(hoststr,UrlSpecHost(spec))) &&
+    (!spec->host || WildcardMatch(hoststr,UrlSpecHost(spec),0)) &&
     (spec->port==-1 || (!portstr && spec->port==0) || (portstr && atoi(portstr)==spec->port)) &&
-    (!spec->path || WildcardMatch(path,UrlSpecPath(spec))) &&
-    (!spec->args || (args && WildcardMatch(args,UrlSpecArgs(spec))) || (!args && *UrlSpecArgs(spec)==0)))
+    (!spec->path || WildcardMatch(path,UrlSpecPath(spec),spec->nocase)) &&
+    (!spec->args || (args && WildcardMatch(args,UrlSpecArgs(spec),spec->nocase)) || (!args && *UrlSpecArgs(spec)==0)))
    {
     match=(spec->proto?strlen(UrlSpecProto(spec)):0)+
           (spec->host ?strlen(UrlSpecHost(spec) ):0)+
@@ -366,42 +349,49 @@ int MatchUrlSpecification(UrlSpec *spec,char *proto,char *host,char *path,char *
   char *string The fixed string that is being matched.
 
   char *pattern The pattern to match against.
+
+  int nocase A flag that if set to 1 ignores the case of the string.
+
+  By Paul A. Rombouts <p.a.rombouts@home.nl>, handles more than two '*' using simpler algorithm than previously.
+
+  See also the strstrn() function at the bottom of this file.
   ++++++++++++++++++++++++++++++++++++++*/
 
-int WildcardMatch(char *string,char *pattern)
+int WildcardMatch(char *string,char *pattern,int nocase)
 {
  int len_beg;
  char *midstr, *endstr;
  char *pattstr, *starp=strchr(pattern,'*');
 
  if(!starp)
-    return(!strcmp(string,pattern));
+    return(( nocase && !strcasecmp(string,pattern)) ||
+           (!nocase && !strcmp(string,pattern)));
 
  len_beg=starp-pattern;
- if(strncmp(string,pattern,len_beg))
+ if(( nocase && strncasecmp(string,pattern,len_beg)) ||
+    (!nocase && strncmp(string,pattern,len_beg)))
     return(0);
 
  midstr=string+len_beg;
 
  while(pattstr=starp+1,starp=strchr(pattstr,'*'))
    {
-    char *match;
-
-    *starp=0;
-    match=strstr(midstr,pattstr);
-    *starp='*';
+    int len_patt=starp-pattstr;
+    char *match=strstrn(midstr,pattstr,len_patt,nocase);
 
     if(!match)
        return(0);
 
-    midstr=match+(starp-pattstr);
+    midstr=match+len_patt;
    }
 
- endstr= strchr(midstr,0)-strlen(pattstr);
+ endstr=midstr+strlen(midstr)-strlen(pattstr);
+
  if(midstr>endstr)
     return(0);
 
- return(!strcmp(endstr,pattstr));
+ return(( nocase && !strcasecmp(endstr,pattstr)) ||
+        (!nocase && !strcmp(endstr,pattstr)));
 }
 
 
@@ -539,6 +529,10 @@ char *ConfigEntryString(ConfigItem item,int which)
 
  /* Send the results back */
 
+ if(url) free(url);
+ if(key) free(key);
+ if(val) free(val);
+
  return(string);
 }
 
@@ -848,6 +842,7 @@ static char* sprintf_url_spec(UrlSpec *urlspec)
  if(urlspec->args ) newlen+=strlen(UrlSpecArgs (urlspec));
 
  string=(char*)malloc(newlen);
+ *string=0;
 
  if(urlspec->negated)
    {
@@ -855,6 +850,12 @@ static char* sprintf_url_spec(UrlSpec *urlspec)
     strpos+=strlen(string+strpos);
    }
 
+ if(urlspec->nocase)
+   {
+    sprintf(string+strpos,"%s","~");
+    strpos+=strlen(string+strpos);
+   }
+
  sprintf(string+strpos,"%s://",urlspec->proto?UrlSpecProto(urlspec):"*");
  strpos+=strlen(string+strpos);
 
@@ -878,3 +879,159 @@ static char* sprintf_url_spec(UrlSpec *urlspec)
 
  return(string);
 }
+
+
+
+/* Return the offset of one string within another.
+   Copyright (C) 1994, 1996, 1997, 2000 Free Software Foundation, Inc.
+
+   The GNU C Library is free software; you can redistribute it and/or
+   modify it under the terms of the GNU Lesser General Public
+   License as published by the Free Software Foundation; either
+   version 2.1 of the License, or (at your option) any later version.
+
+   The GNU C Library is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   Lesser General Public License for more details.
+
+   You should have received a copy of the GNU Lesser General Public
+   License along with the GNU C Library; if not, write to the Free
+   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
+   02111-1307 USA.  */
+
+/*
+ * My personal strstr() implementation that beats most other algorithms.
+ * Until someone tells me otherwise, I assume that this is the
+ * fastest implementation of strstr() in C.
+ * I deliberately chose not to comment it.  You should have at least
+ * as much fun trying to understand it, as I had to write it :-).
+ *
+ * Stephen R. van den Berg, berg@pool.informatik.rwth-aachen.de	*/
+
+/* strstrn() is a variation of strstr() that only tries to find
+   the first n characters of "needle" in "haystack".
+   If strlen(needle) happens to be less than n, strstrn() behaves 
+   exactly like strstr().
+   Modifications made by Paul Rombouts <p.a.rombouts@home.nl>.
+*/
+
+/* Added a parameter "nocase" to select case insensitive test (value 1)
+   or case sensitive (0).
+   Modification by Marc Boucher.
+*/
+
+typedef unsigned chartype;
+
+static char *strstrn(const char *phaystack, const char *pneedle, size_t n, int nocase)
+{
+  register const unsigned char *haystack, *needle;
+  register chartype b, c;
+  const unsigned char *needle_end;
+
+  char *lowhaystack=NULL;
+  char *lowneedle=NULL;
+
+  if (nocase) {
+     int i;
+     lowhaystack=malloc(strlen(phaystack)+1);
+     lowneedle=malloc(strlen(phaystack)+1);
+
+     for(i=0;phaystack[i];i++) { lowhaystack[i]=tolower(phaystack[i]); }
+     lowhaystack[i]=0;
+     for(i=0;pneedle[i];i++) { lowneedle[i]=tolower(pneedle[i]); }
+     lowneedle[i]=0;
+     
+     phaystack=lowhaystack;
+     pneedle=lowneedle;
+     }
+
+  haystack = (const unsigned char *) phaystack;
+  needle = (const unsigned char *) pneedle;
+  needle_end = needle+n;
+
+  if (needle != needle_end && (b = *needle) != '\0' )
+    {
+      haystack--;                               /* possible ANSI violation */
+      do
+        {
+          c = *++haystack;
+          if (c == '\0')
+            goto ret0;
+        }
+      while (c != b);
+
+      if (++needle == needle_end || (c = *needle) == '\0')
+        goto foundneedle;
+      ++needle;
+      goto jin;
+
+      for (;;)
+        {
+          register chartype a;
+          register const unsigned char *rhaystack, *rneedle;
+
+          do
+            {
+              a = *++haystack;
+              if (a == '\0')
+                goto ret0;
+              if (a == b)
+                break;
+              a = *++haystack;
+              if (a == '\0')
+                goto ret0;
+shloop:
+              ;
+            }
+          while (a != b);
+
+jin:      a = *++haystack;
+          if (a == '\0')
+            goto ret0;
+
+          if (a != c)
+            goto shloop;
+
+          rhaystack = haystack-- + 1;
+          if(needle == needle_end) goto foundneedle;
+          rneedle = needle;
+          a = *rneedle;
+
+          if (*rhaystack == a)
+            do
+              {
+                if (a == '\0')
+                  goto foundneedle;
+                ++rhaystack;
+                if(++needle == needle_end) goto foundneedle;
+                a = *needle;
+                if (*rhaystack != a)
+                  break;
+                if (a == '\0')
+                  goto foundneedle;
+                ++rhaystack;
+                if(++needle == needle_end) goto foundneedle;
+                a = *needle;
+              }
+            while (*rhaystack == a);
+
+          needle = rneedle;             /* took the register-poor approach */
+
+          if (a == '\0')
+            break;
+        }
+    }
+foundneedle:
+  if (nocase) {
+     if (lowhaystack) free(lowhaystack);
+     if (lowneedle) free(lowneedle);
+     }
+  return (char*) haystack;
+ret0:
+  if (nocase) {
+     if (lowhaystack) free(lowhaystack);
+     if (lowneedle) free(lowneedle);
+     }
+  return 0;
+}
