@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/ftp.c 1.55 2002/03/02 17:08:09 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/ftp.c 1.71 2004/01/17 16:33:05 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8b.
   Functions for getting URLs using FTP.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1997,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1997,98,99,2000,01,02,03,04 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -20,9 +20,6 @@
 #include <string.h>
 #include <ctype.h>
 
-#include <sys/types.h>
-#include <unistd.h>
-
 #if TIME_WITH_SYS_TIME
 # include <sys/time.h>
 # include <time.h>
@@ -34,12 +31,11 @@
 # endif
 #endif
 
-#include <fcntl.h>
-
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
-#include "config.h"
 #include "errors.h"
+#include "config.h"
 #include "sockets.h"
 #include "proto.h"
 
@@ -48,18 +44,19 @@
 #define DEBUG_FTP 0
 
 /*+ Set to the name of the proxy if there is one. +*/
-static char *proxy=NULL;
+static char /*@null@*/ /*@observer@*/ *proxy=NULL;
 
 /*+ The file descriptor of the socket +*/
 static int server_ctrl=-1,      /*+ for the control connection to the server. +*/
            server_data=-1;      /*+ for the data connection to the server. +*/
 
 /*+ A header to contain the reply. +*/
-static Header *bufferhead;
+static /*@only@*/ /*@null@*/ Header *bufferhead=NULL;
 
-/*+ A buffer to contain the reply +*/
-static char *buffer=NULL,       /*+ body. +*/
-            *buffertail=NULL;   /*+ tail. +*/
+/*+ A buffer to contain the reply body. +*/
+static char /*@only@*/ /*@null@*/ *buffer=NULL;
+/*+ A buffer to contain the reply tail. +*/
+static char /*@only@*/ /*@null@*/ *buffertail=NULL;
 
 /*+ The number of characters in the buffer +*/
 static int nbuffer=0,           /*+ in total for the body part. +*/
@@ -67,7 +64,7 @@ static int nbuffer=0,           /*+ in total for the body part. +*/
            nbuffertail=0,       /*+ in total for the tail . +*/
            nreadtail=0;         /*+ that have been read from the tail. +*/
 
-static char *htmlise_dir_entry(char *line);
+static /*@null@*/ char *htmlise_dir_entry(void);
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -104,10 +101,15 @@ char *FTP_Open(URL *Url)
  /* Open the connection. */
 
  server_ctrl=OpenClientSocket(hoststr,server_port);
- init_buffer(server_ctrl);
 
  if(server_ctrl==-1)
-    msg=PrintMessage(Warning,"Cannot open the FTP control connection to %s port %d; [%!s].",hoststr,server_port);
+    msg=GetPrintMessage(Warning,"Cannot open the FTP control connection to %s port %d; [%!s].",hoststr,server_port);
+ else
+   {
+    init_io(server_ctrl);
+    configure_io_read(server_ctrl,ConfigInteger(SocketTimeout),0,0);
+    configure_io_write(server_ctrl,ConfigInteger(SocketTimeout),0,0);
+   }
 
  RejoinHostPort(server_host,hoststr,portstr);
 
@@ -159,10 +161,10 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
     PrintMessage(ExtraDebug,"Outgoing Request Head (to proxy)\n%s",head);
 
     if(write_string(server_ctrl,head)==-1)
-       msg=PrintMessage(Warning,"Failed to write head to remote FTP proxy; [%!s].");
+       msg=GetPrintMessage(Warning,"Failed to write head to remote FTP proxy; [%!s].");
     if(request_body)
        if(write_data(server_ctrl,request_body->content,request_body->length)==-1)
-          msg=PrintMessage(Warning,"Failed to write body to remote FTP proxy; [%!s].");
+          msg=GetPrintMessage(Warning,"Failed to write body to remote FTP proxy; [%!s].");
 
     free(head);
 
@@ -208,8 +210,9 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: connected; got: %s",str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: connected; got: %s",str);
 
     if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
       {
@@ -229,14 +232,18 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(!str)
    {
-    msg=PrintMessage(Warning,"No reply from FTP server when connected; timed out?");
+    msg=GetPrintMessage(Warning,"No reply from FTP server when connected; timed out?");
+    free(path);
     return(msg);
    }
- else if(atoi(str)!=220)
+
+ if(atoi(str)!=220)
    {
     char *p=str+strlen(str)-1;
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message when connected to FTP server.",str);
+    msg=GetPrintMessage(Warning,"Got '%s' message when connected to FTP server.",str);
+    free(path);
+    free(str);
     return(msg);
    }
 
@@ -244,27 +251,34 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"USER %s\r\n",user)==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'USER' command to remote FTP host; [%!s].");
+    msg=GetPrintMessage(Warning,"Failed to write 'USER' command to remote FTP host; [%!s].");
+    free(path);
+    free(str);
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: sent 'USER %s'; got: %s",user,str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: sent 'USER %s'; got: %s",user,str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
  if(!str)
    {
-    msg=PrintMessage(Warning,"No reply from FTP server to 'USER' command; timed out?");
+    msg=GetPrintMessage(Warning,"No reply from FTP server to 'USER' command; timed out?");
+    free(path);
     return(msg);
    }
- else if(atoi(str)!=230 && atoi(str)!=331)
+
+ if(atoi(str)!=230 && atoi(str)!=331)
    {
     char *p=str+strlen(str)-1;
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message after sending 'USER' command to FTP server.",str);
+    msg=GetPrintMessage(Warning,"Got '%s' message after sending 'USER' command to FTP server.",str);
+    free(path);
+    free(str);
     return(msg);
    }
 
@@ -272,14 +286,17 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    {
     if(write_formatted(server_ctrl,"PASS %s\r\n",pass)==-1)
       {
-       msg=PrintMessage(Warning,"Failed to write 'PASS' command to remote FTP host; [%!s].");
+       msg=GetPrintMessage(Warning,"Failed to write 'PASS' command to remote FTP host; [%!s].");
+       free(path);
+       free(str);
        return(msg);
       }
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str);
-       PrintMessage(ExtraDebug,"FTP: sent 'PASS %s'; got: %s",pass,str);
+       str=read_line(server_ctrl,str);
+       if(str)
+          PrintMessage(ExtraDebug,"FTP: sent 'PASS %s'; got: %s",pass,str);
 
        if(!file && !*path && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
          {
@@ -299,14 +316,25 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
     if(!str)
       {
-       msg=PrintMessage(Warning,"No reply from FTP server to 'PASS' command; timed out?");
+       msg=GetPrintMessage(Warning,"No reply from FTP server to 'PASS' command; timed out?");
+       free(path);
        return(msg);
       }
-    else if(atoi(str)!=202 && atoi(str)!=230)
+
+    if(atoi(str)==530)
+      {
+       bufferhead=CreateHeader("HTTP/1.0 401 FTP invalid password",0);
+
+       goto near_end;
+      }
+
+    if(atoi(str)!=202 && atoi(str)!=230)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'PASS' command to FTP server.",str);
+       msg=GetPrintMessage(Warning,"Got '%s' message after sending 'PASS' command to FTP server.",str);
+       free(path);
+       free(str);
        return(msg);
       }
    }
@@ -315,14 +343,17 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"CWD %s\r\n",*path?path:"/")==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
+    msg=GetPrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
+    free(path);
+    free(str);
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",*path?path:"/",str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",*path?path:"/",str);
 
     if(!file && str && isdigit(str[0]) && isdigit(str[1]) && isdigit(str[2]) && str[3]=='-')
       {
@@ -342,14 +373,18 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(!str)
    {
-    msg=PrintMessage(Warning,"No reply from FTP server to 'CWD' command; timed out?");
+    msg=GetPrintMessage(Warning,"No reply from FTP server to 'CWD' command; timed out?");
+    free(path);
     return(msg);
    }
- else if(atoi(str)!=250)
+
+ if(atoi(str)!=250)
    {
     char *p=str+strlen(str)-1;
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message after sending 'CWD' command to FTP server.",str);
+    msg=GetPrintMessage(Warning,"Got '%s' message after sending 'CWD' command to FTP server.",str);
+    free(path);
+    free(str);
     return(msg);
    }
 
@@ -359,43 +394,50 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    {
     if(write_formatted(server_ctrl,"CWD %s\r\n",file)==-1)
       {
-       msg=PrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
+       msg=GetPrintMessage(Warning,"Failed to write 'CWD' command to remote FTP host; [%!s].");
+       free(path);
+       free(str);
        return(msg);
       }
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str);
-       PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",file,str);
+       str=read_line(server_ctrl,str);
+       if(str)
+          PrintMessage(ExtraDebug,"FTP: sent 'CWD %s' got: %s",file,str);
       }
     while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
     if(!str)
       {
-       msg=PrintMessage(Warning,"No reply from FTP server to 'CWD' command; timed out?");
+       msg=GetPrintMessage(Warning,"No reply from FTP server to 'CWD' command; timed out?");
+       free(path);
        return(msg);
       }
-    else if(atoi(str)!=250 && atoi(str)!=501 && atoi(str)!=530 && atoi(str)!=550)
+
+    if(atoi(str)!=250 && atoi(str)!=501 && atoi(str)!=530 && atoi(str)!=550)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'CWD' command to FTP server.",str);
+       msg=GetPrintMessage(Warning,"Got '%s' message after sending 'CWD' command to FTP server.",str);
+       free(path);
+       free(str);
        return(msg);
       }
 
     if(atoi(str)==250)
       {
-       char *loc=(char*)malloc(strlen(Url->link)+2);
+       char *loc=(char*)malloc(strlen(Url->file)+2);
 
-       strcpy(loc,Url->link);
+       strcpy(loc,Url->file);
        strcat(loc,"/");
 
        bufferhead=CreateHeader("HTTP/1.0 302 FTP is a directory",0);
        AddToHeader(bufferhead,"Location",loc);
 
-       buffer=HTMLMessageBody(-1,"FTPDirRedirect",
-                              "url",Url->link,
-                              NULL);
+       buffer=HTMLMessageString("Redirect",
+                                "location",loc,
+                                NULL);
 
        free(loc);
 
@@ -407,27 +449,34 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_formatted(server_ctrl,"TYPE %c\r\n",file?'I':'A')==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'TYPE' command to remote FTP host; [%!s].");
+    msg=GetPrintMessage(Warning,"Failed to write 'TYPE' command to remote FTP host; [%!s].");
+    free(path);
+    free(str);
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: sent 'TYPE %c'; got: %s",file?'I':'A',str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: sent 'TYPE %c'; got: %s",file?'I':'A',str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
  if(!str)
    {
-    msg=PrintMessage(Warning,"No reply from FTP server to 'TYPE' command; timed out?");
+    msg=GetPrintMessage(Warning,"No reply from FTP server to 'TYPE' command; timed out?");
+    free(path);
     return(msg);
    }
- else if(atoi(str)!=200)
+
+ if(atoi(str)!=200)
    {
     char *p=str+strlen(str)-1;
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message after sending 'TYPE' command to FTP server.",str);
+    msg=GetPrintMessage(Warning,"Got '%s' message after sending 'TYPE' command to FTP server.",str);
+    free(path);
+    free(str);
     return(msg);
    }
 
@@ -439,20 +488,24 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        if(write_formatted(server_ctrl,"SIZE %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'SIZE' command to remote FTP host; [%!s].");
+          msg=GetPrintMessage(Warning,"Failed to write 'SIZE' command to remote FTP host; [%!s].");
+          free(path);
+          free(str);
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str);
-          PrintMessage(ExtraDebug,"FTP: sent 'SIZE %s'; got: %s",file,str);
+          str=read_line(server_ctrl,str);
+          if(str)
+             PrintMessage(ExtraDebug,"FTP: sent 'SIZE %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
        if(!str)
          {
-          msg=PrintMessage(Warning,"No reply from FTP server to 'SIZE' command; timed out?");
+          msg=GetPrintMessage(Warning,"No reply from FTP server to 'SIZE' command; timed out?");
+          free(path);
           return(msg);
          }
 
@@ -462,20 +515,24 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
        if(write_formatted(server_ctrl,"MDTM %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'MDTM' command to remote FTP host; [%!s].");
+          msg=GetPrintMessage(Warning,"Failed to write 'MDTM' command to remote FTP host; [%!s].");
+          free(path);
+          free(str);
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str);
-          PrintMessage(ExtraDebug,"FTP: sent 'MDTM %s'; got: %s",file,str);
+          str=read_line(server_ctrl,str);
+          if(str)
+             PrintMessage(ExtraDebug,"FTP: sent 'MDTM %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
        if(!str)
          {
-          msg=PrintMessage(Warning,"No reply from FTP server to 'MDTM' command; timed out?");
+          msg=GetPrintMessage(Warning,"No reply from FTP server to 'MDTM' command; timed out?");
+          free(path);
           return(msg);
          }
 
@@ -505,20 +562,24 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
  if(write_string(server_ctrl,"EPSV\r\n")==-1)
    {
-    msg=PrintMessage(Warning,"Failed to write 'EPSV' command to remote FTP host; [%!s].");
+    msg=GetPrintMessage(Warning,"Failed to write 'EPSV' command to remote FTP host; [%!s].");
+    free(path);
+    free(str);
     return(msg);
    }
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: sent 'EPSV'; got: %s",str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: sent 'EPSV'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
  if(!str)
    {
-    msg=PrintMessage(Warning,"No reply from FTP server to 'EPSV' command; timed out?");
+    msg=GetPrintMessage(Warning,"No reply from FTP server to 'EPSV' command; timed out?");
+    free(path);
     return(msg);
    }
 
@@ -529,12 +590,16 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'EPSV' command, cannot parse.",str);
+       msg=GetPrintMessage(Warning,"Got '%s' message after sending 'EPSV' command, cannot parse.",str);
+       free(path);
+       free(str);
        return(msg);
       }
     if(SocketRemoteName(server_ctrl,&host,NULL,NULL))
       {
-       msg=PrintMessage(Warning,"Cannot determine server address.");
+       msg=GetPrintMessage(Warning,"Cannot determine server address.");
+       free(path);
+       free(str);
        return(msg);
       }
    }
@@ -543,7 +608,9 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
     char *p=str+strlen(str)-1;
 
     while(*p=='\n' || *p=='\r') *p--=0;
-    msg=PrintMessage(Warning,"Got '%s' message after sending 'EPSV' command",str);
+    msg=GetPrintMessage(Warning,"Got '%s' message after sending 'EPSV' command",str);
+    free(path);
+    free(str);
     return(msg);
    }
  else
@@ -554,27 +621,34 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
     if(write_string(server_ctrl,"PASV\r\n")==-1)
       {
-       msg=PrintMessage(Warning,"Failed to write 'PASV' command to remote FTP host; [%!s].");
+       msg=GetPrintMessage(Warning,"Failed to write 'PASV' command to remote FTP host; [%!s].");
+       free(path);
+       free(str);
        return(msg);
       }
 
     do
       {
-       str=read_line_or_timeout(server_ctrl,str);
-       PrintMessage(ExtraDebug,"FTP: sent 'PASV'; got: %s",str);
+       str=read_line(server_ctrl,str);
+       if(str)
+          PrintMessage(ExtraDebug,"FTP: sent 'PASV'; got: %s",str);
       }
     while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
     if(!str)
       {
-       msg=PrintMessage(Warning,"No reply from FTP server to 'PASV' command; timed out?");
+       msg=GetPrintMessage(Warning,"No reply from FTP server to 'PASV' command; timed out?");
+       free(path);
        return(msg);
       }
-    else if(atoi(str)!=227)
+
+    if(atoi(str)!=227)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command",str);
+       msg=GetPrintMessage(Warning,"Got '%s' message after sending 'PASV' command",str);
+       free(path);
+       free(str);
        return(msg);
       }
 
@@ -588,7 +662,9 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        char *p=str+strlen(str)-1;
        while(*p=='\n' || *p=='\r') *p--=0;
-       msg=PrintMessage(Warning,"Got '%s' message after sending 'PASV' command, cannot parse.",str);
+       msg=GetPrintMessage(Warning,"Got '%s' message after sending 'PASV' command, cannot parse.",str);
+       free(path);
+       free(str);
        return(msg);
       }
 
@@ -601,13 +677,20 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
    }
 
  server_data=OpenClientSocket(host,port);
- init_buffer(server_data);
 
  if(server_data==-1)
    {
-    msg=PrintMessage(Warning,"Cannot open the FTP data connection [%!s].");
+    msg=GetPrintMessage(Warning,"Cannot open the FTP data connection [%!s].");
+    free(path);
+    free(str);
     return(msg);
    }
+ else
+   {
+    init_io(server_data);
+    configure_io_read(server_data,ConfigInteger(SocketTimeout),0,0);
+    configure_io_write(server_data,ConfigInteger(SocketTimeout),0,0);
+   }
 
  /* Make the request */
 
@@ -621,14 +704,17 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
        if(write_formatted(server_ctrl,"RETR %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host; [%!s].");
+          msg=GetPrintMessage(Warning,"Failed to write 'RETR' command to remote FTP host; [%!s].");
+          free(path);
+          free(str);
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str);
-          PrintMessage(ExtraDebug,"FTP: sent 'RETR %s'; got: %s",file,str);
+          str=read_line(server_ctrl,str);
+          if(str)
+             PrintMessage(ExtraDebug,"FTP: sent 'RETR %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -647,14 +733,17 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
 
           if(write_formatted(server_ctrl,"%s\r\n",command)==-1)
             {
-             msg=PrintMessage(Warning,"Failed to write '%s' command to remote FTP host; [%!s].",command);
+             msg=GetPrintMessage(Warning,"Failed to write '%s' command to remote FTP host; [%!s].",command);
+             free(path);
+             free(str);
              return(msg);
             }
 
           do
             {
-             str=read_line_or_timeout(server_ctrl,str);
-             PrintMessage(ExtraDebug,"FTP: sent '%s'; got: %s",command,str);
+             str=read_line(server_ctrl,str);
+             if(str)
+                PrintMessage(ExtraDebug,"FTP: sent '%s'; got: %s",command,str);
             }
           while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -662,7 +751,7 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
             {
              char *p=str+strlen(str)-1;
              while(*p=='\n' || *p=='\r') *p--=0;
-             msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
+             PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
             }
           else
              break;
@@ -684,10 +773,12 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
          {
           char *p=str+strlen(str)-1;
           while(*p=='\n' || *p=='\r') *p--=0;
-          msg=PrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
+          msg=GetPrintMessage(Warning,"Got '%s' message after sending '%s' command to FTP server.",str,command);
+          free(str);
          }
        else
-          msg=PrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",command);
+          msg=GetPrintMessage(Warning,"No reply from FTP server to '%s' command; timed out?",command);
+       free(path);
        return(msg);
       }
    }
@@ -697,14 +788,17 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       {
        if(write_formatted(server_ctrl,"STOR %s\r\n",file)==-1)
          {
-          msg=PrintMessage(Warning,"Failed to write 'STOR' command to remote FTP host; [%!s].");
+          msg=GetPrintMessage(Warning,"Failed to write 'STOR' command to remote FTP host; [%!s].");
+          free(path);
+          free(str);
           return(msg);
          }
 
        do
          {
-          str=read_line_or_timeout(server_ctrl,str);
-          PrintMessage(ExtraDebug,"FTP: sent 'STOR %s'; got: %s",file,str);
+          str=read_line(server_ctrl,str);
+          if(str)
+             PrintMessage(ExtraDebug,"FTP: sent 'STOR %s'; got: %s",file,str);
          }
        while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
@@ -712,7 +806,9 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
       }
     else
       {
-       msg=PrintMessage(Warning,"Cannot use the PUT method on a directory name");
+       msg=GetPrintMessage(Warning,"Cannot use the PUT method on a directory name");
+       free(path);
+       free(str);
        return(msg);
       }
 
@@ -729,10 +825,12 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
          {
           char *p=str+strlen(str)-1;
           while(*p=='\n' || *p=='\r') *p--=0;
-          msg=PrintMessage(Warning,"Got '%s' message after sending 'STOR' command to FTP server.",str);
+          msg=GetPrintMessage(Warning,"Got '%s' message after sending 'STOR' command to FTP server.",str);
+          free(str);
          }
        else
-          msg=PrintMessage(Warning,"No reply from FTP server to 'STOR' command; timed out?");
+          msg=GetPrintMessage(Warning,"No reply from FTP server to 'STOR' command; timed out?");
+       free(path);
        return(msg);
       }
 
@@ -763,22 +861,16 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
          {
           time_t lastmodtime=DateToTimeT(ims);
 
-          if(modtime<lastmodtime)
+          if(modtime<=lastmodtime)
             {
              bufferhead->status=304;
              RemoveFromHeader(bufferhead,"Content-Length");
 
-             CloseSocket(server_data);
-             server_data=-1;
-
              goto near_end;
             }
          }
       }
-   }
 
- if(!strcmp(request_head->method,"GET"))
-   {
     if(!file)
       {
        if(msg_reply)
@@ -788,19 +880,20 @@ char *FTP_Request(URL *Url,Header *request_head,Body *request_body)
           free(old_msg_reply);
          }
 
-       buffer=HTMLMessageBody(-1,"FTPDir-Head",
-                              "url",Url->name,
-                              "message",msg_reply,
-                              NULL);
+       buffer=HTMLMessageString("FTPDir-Head",
+                                "url",Url->name,
+                                "base",Url->file,
+                                "message",msg_reply,
+                                NULL);
 
-       buffertail=HTMLMessageBody(-1,"FTPDir-Tail",
-                                  NULL);
+       buffertail=HTMLMessageString("FTPDir-Tail",
+                                    NULL);
       }
    }
  else
-    buffer=HTMLMessageBody(-1,"FTPPut",
-                           "url",Url->name,
-                           NULL);
+    buffer=HTMLMessageString("FTPPut",
+                             "url",Url->name,
+                             NULL);
 
 near_end:
 
@@ -810,8 +903,8 @@ near_end:
  nbuffer=buffer?strlen(buffer):0; nread=0;
  nbuffertail=buffertail?strlen(buffertail):0; nreadtail=0;
 
- free(str);
  free(path);
+ free(str);
 
  return(msg);
 }
@@ -861,7 +954,7 @@ int FTP_ReadBody(char *s,int n)
  /* Take a simple route if it is proxied. */
 
  if(proxy)
-    return(read_data_or_timeout(server_ctrl,s,n));
+    return(read_data(server_ctrl,s,n));
  
  /* Else send the data then the tail. */
 
@@ -874,7 +967,7 @@ int FTP_ReadBody(char *s,int n)
        s[m]=buffertail[nreadtail];
    }
  else if(!buffer && !buffertail) /* File not dir entry. */
-    m=read_data_or_timeout(server_data,s,n);
+    m=read_data(server_data,s,n);
  else if(buffer && buffertail)  /* Middle of dir entry */
    {
     for(;nread<nbuffer && m<n;nread++,m++)
@@ -882,7 +975,8 @@ int FTP_ReadBody(char *s,int n)
 
     if(nread==nbuffer)
       {
-       buffer=htmlise_dir_entry(buffer);
+       free(buffer);
+       buffer=htmlise_dir_entry();
        if(buffer)
           nbuffer=strlen(buffer),nread=0;
       }
@@ -916,14 +1010,15 @@ int FTP_Close(void)
 {
  int err=0;
  char *str=NULL;
+ unsigned long r1=0,w1=0,r2=0,w2=0;
 
  /* Take a simple route if it is proxied. */
 
  if(proxy)
    {
-#if USE_ZLIB
-    finish_zlib_buffer(server_ctrl);
-#endif
+    finish_tell_io(server_ctrl,&r1,&w1);
+
+    PrintMessage(Inform,"Server bytes; %d Read, %d Written.",r1,w1); /* Used in audit-usage.pl */
 
     return(CloseSocket(server_ctrl));
    }
@@ -931,22 +1026,31 @@ int FTP_Close(void)
  /* Else say goodbye and close all of the sockets, */
 
  if(server_data!=-1)
+   {
+    finish_tell_io(server_data,&r2,&w2);
     CloseSocket(server_data);
+   }
 
  write_string(server_ctrl,"QUIT\r\n");
 
  do
    {
-    str=read_line_or_timeout(server_ctrl,str);
-    PrintMessage(ExtraDebug,"FTP: sent 'QUIT'; got: %s",str);
+    str=read_line(server_ctrl,str);
+    if(str)
+       PrintMessage(ExtraDebug,"FTP: sent 'QUIT'; got: %s",str);
    }
  while(str && (!isdigit(str[0]) || !isdigit(str[1]) || !isdigit(str[2]) || str[3]!=' '));
 
+ finish_tell_io(server_ctrl,&r2,&w2);
  err=CloseSocket(server_ctrl);
 
+ PrintMessage(Inform,"Server bytes; %d Read, %d Written.",r1+r2,w1+w2); /* Used in audit-usage.pl */
+
+ if(str)
+    free(str);
+
  if(buffer)
     free(buffer);
-
  if(buffertail)
     free(buffertail);
 
@@ -958,21 +1062,21 @@ int FTP_Close(void)
   Convert a line from the ftp server dir listing into a pretty listing.
 
   char *htmlise_dir_entry Returns the next line.
-
-  char *line The previous line.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static char *htmlise_dir_entry(char *line)
+static char *htmlise_dir_entry(void)
 {
  int i,isdir=0,islink=0;
  char *q,*p[16],*l;
  int file=0,link=0;
 
- l=read_line_or_timeout(server_data,line);
+ l=read_line(server_data,NULL);
 
  if(!l)
     return(l);
 
+ p[0]=l;
+
  for(q=l,i=0;*q && i<16;i++)
    {
     while(*q && isspace(*q))
