@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/src/RCS/purge.c 2.51 2002/03/24 16:33:55 amb Exp $
+  $Header: /home/amb/wwwoffle/src/RCS/purge.c 2.60 2003/03/16 09:59:48 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.7a.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.8.
   Purge old files from the cache.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98,99,2000,01,02 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01,02,03 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -75,10 +75,11 @@
 #endif
 
 #include "wwwoffle.h"
+#include "io.h"
 #include "misc.h"
-#include "proto.h"
-#include "config.h"
 #include "errors.h"
+#include "config.h"
+#include "proto.h"
 
 
 /*+ Need this for Win32 to use binary mode +*/
@@ -87,16 +88,12 @@
 #endif
 
 
-/*+ The file descriptor of the sppol directory. +*/
-extern int fSpoolDir;
-
-
 /* Local functions */
 
-static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress_age,long *remain,long *deleted,long *compressed);
-static void what_purge_compress_age(char *proto,char *host,char *file,int *purge_age,int *compress_age);
+static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress_age,/*@out@*/ unsigned long *remain,/*@out@*/ unsigned long *deleted,/*@out@*/ unsigned long *compressed);
+static void what_purge_compress_age(char *proto,char *host,/*@null@*/ char *file,/*@out@*/ int *purge_age,/*@out@*/ int *compress_age);
 #if USE_ZLIB
-static long compress_file(char *proto,char *host,char *file);
+static unsigned long compress_file(char *proto,char *host,char *file);
 #endif
 
 /*+ Set this to 0 for debugging so that nothing is deleted. +*/
@@ -121,6 +118,8 @@ static int pass;
 static unsigned long blocksize;
 
 /*+ The configuraion file options, looked up once then used. +*/
+static int purge_max_size;
+static int purge_min_free;
 static int purge_use_mtime;
 static int purge_use_url;
 static int purge_default_age;
@@ -149,9 +148,11 @@ void PurgeCache(int fd)
  struct STATFS sbuf;
  struct stat buf;
 
- what_purge_compress_age("*","*",NULL,&purge_default_age,&compress_default_age);
+ purge_max_size=ConfigInteger(PurgeMaxSize);
+ purge_min_free=ConfigInteger(PurgeMinFree);
  purge_use_mtime=ConfigBoolean(PurgeUseMTime);
  purge_use_url=ConfigBoolean(PurgeUseURL);
+ what_purge_compress_age("*","*",NULL,&purge_default_age,&compress_default_age);
 
  now=time(NULL)+600;
 
@@ -210,7 +211,7 @@ void PurgeCache(int fd)
 
     write_string(fd,"\n");
 
-    if(ConfigInteger(PurgeCacheSize))
+    if(purge_max_size>0 || purge_min_free>0)
        if(pass==1)
           write_string(fd,"Pass 1: Checking dates and sizes of files:\n");
        else
@@ -220,12 +221,12 @@ void PurgeCache(int fd)
 
     if(pass==1)
       {
-       if(ConfigBoolean(PurgeUseMTime))
+       if(purge_use_mtime)
           write_string(fd,"  (Using modification time.)\n");
        else
           write_string(fd,"  (Using last access time.)\n");
 
-       if(ConfigBoolean(PurgeUseURL))
+       if(purge_use_url)
           write_string(fd,"  (Using the full URL.)\n");
        else
           write_string(fd,"  (Using the hostname and protocol only.)\n");
@@ -255,11 +256,11 @@ void PurgeCache(int fd)
 
        dir=opendir(".");
        if(!dir)
-         {PrintMessage(Warning,"Cannot open directory '%s' [%!s]; not purged.",proto);fchdir(fSpoolDir);continue;}
+         {PrintMessage(Warning,"Cannot open directory '%s' [%!s]; not purged.",proto);ChangeBackToSpoolDir();continue;}
 
        ent=readdir(dir);  /* skip .  */
        if(!ent)
-         {PrintMessage(Warning,"Cannot read directory '%s' [%!s]; not purged.",proto);closedir(dir);fchdir(fSpoolDir);continue;}
+         {PrintMessage(Warning,"Cannot read directory '%s' [%!s]; not purged.",proto);closedir(dir);ChangeBackToSpoolDir();continue;}
        ent=readdir(dir);  /* skip .. */
 
        /* Search through all of the sub directories. */
@@ -272,7 +273,7 @@ void PurgeCache(int fd)
              PrintMessage(Inform,"Cannot stat directory '%s/%s' [%!s]; race condition?",proto,ent->d_name);
           else if(S_ISDIR(buf.st_mode))
             {
-             long file_blocks,del_blocks,compress_blocks;
+             unsigned long file_blocks,del_blocks,compress_blocks;
              int purge_age,compress_age;
 
              what_purge_compress_age(proto,ent->d_name,NULL,&purge_age,&compress_age);
@@ -309,7 +310,7 @@ void PurgeCache(int fd)
              if(purge_use_url)
                {
                 if(file_blocks==0)
-                   write_formatted(fd,"Purged %6s://%-40s ; empty    (%4ld kB del)\n",proto,ent->d_name,
+                   write_formatted(fd,"Purged %6s://%-40s ; empty    (%4lu kB del)\n",proto,ent->d_name,
                                    Blocks_to_kB(del_blocks+dir_blocks));
                 else
                   {
@@ -317,20 +318,20 @@ void PurgeCache(int fd)
                    if(compress_blocks)
                      {
                       if(del_blocks)
-                         write_formatted(fd,"Purged %6s://%-40s ; %5ld kB (%4ld kB del) (%4ld kB compr)\n",proto,ent->d_name,
+                         write_formatted(fd,"Purged %6s://%-40s ; %5lu kB (%4lu kB del) (%4lu kB compr)\n",proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(del_blocks),Blocks_to_kB(compress_blocks));
                       else
-                         write_formatted(fd,"Purged %6s://%-40s ; %5ld kB               (%4ld kB compr)\n",proto,ent->d_name,
+                         write_formatted(fd,"Purged %6s://%-40s ; %5lu kB               (%4lu kB compr)\n",proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(compress_blocks));
                      }
                    else
 #endif
                      {
                       if(del_blocks)
-                         write_formatted(fd,"Purged %6s://%-40s ; %5ld kB (%4ld kB del)\n",proto,ent->d_name,
+                         write_formatted(fd,"Purged %6s://%-40s ; %5lu kB (%4lu kB del)\n",proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(del_blocks));
                       else
-                         write_formatted(fd,"Purged %6s://%-40s ; %5ld kB\n",proto,ent->d_name,
+                         write_formatted(fd,"Purged %6s://%-40s ; %5lu kB\n",proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks));
                      }
                   }
@@ -344,15 +345,15 @@ void PurgeCache(int fd)
                   {
 #if USE_ZLIB
                    if(compress_blocks)
-                      write_formatted(fd,"Not Purged       %6s://%-40s ; %4ld kB               (%4ld kB compr)\n",proto,ent->d_name,
+                      write_formatted(fd,"Not Purged       %6s://%-40s ; %4lu kB               (%4lu kB compr)\n",proto,ent->d_name,
                                       Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(compress_blocks));
                    else
 #endif
-                      write_formatted(fd,"Not Purged       %6s://%-40s ; %4ld kB\n",proto,ent->d_name,
+                      write_formatted(fd,"Not Purged       %6s://%-40s ; %4lu kB\n",proto,ent->d_name,
                                       Blocks_to_kB(file_blocks+dir_blocks));
                   }
                 else if(file_blocks==0)
-                   write_formatted(fd,"Purged (%2d days) %6s://%-40s ; empty    (%4ld kB del)\n",purge_age,proto,ent->d_name,
+                   write_formatted(fd,"Purged (%2d days) %6s://%-40s ; empty    (%4lu kB del)\n",purge_age,proto,ent->d_name,
                                    Blocks_to_kB(del_blocks+dir_blocks));
                 else
                   {
@@ -360,20 +361,20 @@ void PurgeCache(int fd)
                    if(compress_blocks)
                      {
                       if(del_blocks)
-                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5ld kB (%4ld kB del) (%4ld kB compr)\n",purge_age,proto,ent->d_name,
+                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5lu kB (%4lu kB del) (%4lu kB compr)\n",purge_age,proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(del_blocks),Blocks_to_kB(compress_blocks));
                       else
-                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5ld kB               (%4ld kB compr)\n",purge_age,proto,ent->d_name,
+                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5lu kB               (%4lu kB compr)\n",purge_age,proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(compress_blocks));
                      }
                    else
 #endif
                      {
                       if(del_blocks)
-                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5ld kB (%4ld kB del)\n",purge_age,proto,ent->d_name,
+                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5lu kB (%4lu kB del)\n",purge_age,proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks),Blocks_to_kB(del_blocks));
                       else
-                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5ld kB\n",purge_age,proto,ent->d_name,
+                         write_formatted(fd,"Purged (%2d days) %6s://%-40s ; %5lu kB\n",purge_age,proto,ent->d_name,
                                          Blocks_to_kB(file_blocks+dir_blocks));
                      }
                   }
@@ -394,15 +395,15 @@ void PurgeCache(int fd)
          }
 
        closedir(dir);
-       fchdir(fSpoolDir);
+       ChangeBackToSpoolDir();
       }
 
     write_string(fd,"\n");
 #if USE_ZLIB
-    write_formatted(fd,"Total of %ld directories ; %lu kB (%lu kB deleted) (%lu kB compressed)\n",total_dirs,
+    write_formatted(fd,"Total of %lu directories ; %lu kB (%lu kB deleted) (%lu kB compressed)\n",total_dirs,
                     Blocks_to_kB(total_file_blocks),Blocks_to_kB(total_del_blocks),Blocks_to_kB(total_compress_blocks));
 #else
-    write_formatted(fd,"Total of %ld directories ; %lu kB (%lu kB deleted)\n",total_dirs,
+    write_formatted(fd,"Total of %lu directories ; %lu kB (%lu kB deleted)\n",total_dirs,
                     Blocks_to_kB(total_file_blocks),Blocks_to_kB(total_del_blocks));
 #endif
 
@@ -445,7 +446,7 @@ void PurgeCache(int fd)
        write_formatted(fd,"  (All ages scaled to the range 0 -> %d (default age).)\n",purge_default_age);
        write_string(fd,"\n");
 
-       write_formatted(fd,"Total not purged   ; %5ld kB (%6ld kB free)\n",
+       write_formatted(fd,"Total not purged   ; %5lu kB (%6lu kB free)\n",
                        Blocks_to_kB(age_blocks_used),Blocks_to_kB(age_blocks_free));
        write_string(fd,"\n");
 
@@ -454,28 +455,28 @@ void PurgeCache(int fd)
           age_blocks_used+=blocks_by_age[i];
           age_blocks_free-=blocks_by_age[i];
 
-          if(ConfigInteger(PurgeCacheSize) && age_for_size<0 &&
-             Blocks_to_kB(age_blocks_used)>(1024*ConfigInteger(PurgeCacheSize)))
+          if(purge_max_size>0 && age_for_size<0 &&
+             Blocks_to_kB(age_blocks_used)>(1024*purge_max_size))
             {
              age_for_size=i;
 
-             write_formatted(fd,"Cutoff Age is %2d days for %3d MB cache size\n",age_for_size,ConfigInteger(PurgeCacheSize));
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB cache size\n",age_for_size,purge_max_size);
             }
 
-          if(ConfigInteger(PurgeDiskFree) && diskfree && age_for_free<0 &&
-             Blocks_to_kB(age_blocks_free)<(1024*ConfigInteger(PurgeDiskFree)))
+          if(purge_min_free>0 && diskfree && age_for_free<0 &&
+             Blocks_to_kB(age_blocks_free)<(1024*purge_min_free))
             {
              age_for_free=i;
 
-             write_formatted(fd,"Cutoff Age is %2d days for %3d MB disk free\n",age_for_free,ConfigInteger(PurgeDiskFree));
+             write_formatted(fd,"Cutoff Age is %2d days for %3d MB disk free\n",age_for_free,purge_min_free);
             }
 
           if(i==purge_default_age)
-             write_formatted(fd,"Total all ages     ; %5ld kB (%6ld kB free)\n",
+             write_formatted(fd,"Total all ages     ; %5lu kB (%6lu kB free)\n",
                              Blocks_to_kB(age_blocks_used-blocks_by_age[purge_default_age+1]),
                              diskfree?Blocks_to_kB(age_blocks_free):0);
           else
-             write_formatted(fd,"Newer than %2d day%c ; %5ld kB (%6ld kB free)\n",i+1,i?'s':' ',
+             write_formatted(fd,"Newer than %2d day%c ; %5lu kB (%6lu kB free)\n",i+1,i?'s':' ',
                              Blocks_to_kB(age_blocks_used-blocks_by_age[purge_default_age+1]),
                              diskfree?Blocks_to_kB(age_blocks_free):0);
          }
@@ -507,6 +508,7 @@ void PurgeCache(int fd)
    {
     DIR *dir;
     struct dirent* ent;
+    int count1=0,count2=0;
 
     dir=opendir(".");
     if(!dir)
@@ -521,10 +523,31 @@ void PurgeCache(int fd)
           ent=readdir(dir);  /* skip .. */
 
           while((ent=readdir(dir)))
-             if(!strncmp(ent->d_name,"tmp.",4))
-               {
-                struct stat buf;
+            {
+             struct stat buf;
 
+             if(*ent->d_name=='U' || *ent->d_name=='O')
+               {
+                int s;
+
+                *ent->d_name^='U'^'O';
+                s=stat(ent->d_name,&buf);
+                *ent->d_name^='U'^'O';
+
+                if(s)
+                  {
+                   PrintMessage(Inform,"Outgoing file 'outgoing/%s' is not complete (U* and O* files); deleting it.",ent->d_name);
+#if DO_DELETE
+                   if(unlink(ent->d_name))
+                      PrintMessage(Warning,"Cannot unlink file 'outgoing/%s' [%!s]; race condition?",ent->d_name);
+#else
+                   PrintMessage(Debug,"unlink(outgoing/%s).",ent->d_name);
+#endif
+                   count1++;
+                  }
+               }
+             else if(!strncmp(ent->d_name,"tmp.",4))
+               {
                 if(!stat(ent->d_name,&buf) && buf.st_mtime<(now-60))
                   {
 #if DO_DELETE
@@ -533,15 +556,22 @@ void PurgeCache(int fd)
 #else
                    PrintMessage(Debug,"unlink(outgoing/%s).",ent->d_name);
 #endif
+                   count2++;
                   }
                }
+            }
          }
 
        closedir(dir);
       }
+
+    if(count1)
+       write_formatted(fd,"\nDeleted %d unmatched file (missing O* or U*) from directory 'outgoing'\n\n",count1);
+    if(count2)
+       write_formatted(fd,"\nDeleted %d temporary files from directory 'outgoing'\n\n",count2);
    }
 
- fchdir(fSpoolDir);
+ ChangeBackToSpoolDir();
 
  /* Purge the tmp.* files in temp. */
 
@@ -551,6 +581,7 @@ void PurgeCache(int fd)
    {
     DIR *dir;
     struct dirent* ent;
+    int count=0;
 
     dir=opendir(".");
     if(!dir)
@@ -577,15 +608,19 @@ void PurgeCache(int fd)
 #else
                    PrintMessage(Debug,"unlink(temp/%s).",ent->d_name);
 #endif
+                   count++;
                   }
                }
          }
 
        closedir(dir);
       }
+
+    if(count)
+       write_formatted(fd,"\nDeleted %d temporary files from directory 'temp'.\n\n",count);
    }
 
- fchdir(fSpoolDir);
+ ChangeBackToSpoolDir();
 }
 
 
@@ -600,14 +635,14 @@ void PurgeCache(int fd)
 
   int def_compress age The default compress age to use for this host.
 
-  long *remain Returns the number of blocks in files that are left.
+  unsigned long *remain Returns the number of blocks in files that are left.
 
-  long *deleted Returns the number of blocks in files that were deleted.
+  unsigned long *deleted Returns the number of blocks in files that were deleted.
 
-  long *compressed Returns the number of blocks in files that were compressed.
+  unsigned long *compressed Returns the number of blocks in files that were compressed.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress_age,long *remain,long *deleted,long *compressed)
+static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress_age,unsigned long *remain,unsigned long *deleted,unsigned long *compressed)
 {
  DIR *dir;
  struct dirent* ent;
@@ -623,11 +658,11 @@ static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress
 
  dir=opendir(".");
  if(!dir)
-   {PrintMessage(Warning,"Cannot open directory '%s/%s' [%!s]; not purged.",proto,host);fchdir(fSpoolDir);chdir(proto);return;}
+   {PrintMessage(Warning,"Cannot open directory '%s/%s' [%!s]; not purged.",proto,host);ChangeBackToSpoolDir();chdir(proto);return;}
 
  ent=readdir(dir);  /* skip .  */
  if(!ent)
-   {PrintMessage(Warning,"Cannot read directory '%s/%s' [%!s]; not purged.",proto,host);closedir(dir);fchdir(fSpoolDir);chdir(proto);return;}
+   {PrintMessage(Warning,"Cannot read directory '%s/%s' [%!s]; not purged.",proto,host);closedir(dir);ChangeBackToSpoolDir();chdir(proto);return;}
  ent=readdir(dir);  /* skip .. */
 
  /* Check all of the files for age, and delete as needed. */
@@ -693,7 +728,7 @@ static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress
 
        if(purge_age==-1 || (now-filetime)<(purge_age*(24*3600)))
          {
-          long size;
+          unsigned long size;
 
 #if USE_ZLIB
           if(compress_age!=-1 && (now-filetime)>(compress_age*(24*3600)) && *ent->d_name=='D')
@@ -756,7 +791,7 @@ static void PurgeFiles(char *proto,char *host,int def_purge_age,int def_compress
    }
 
  closedir(dir);
- fchdir(fSpoolDir);
+ ChangeBackToSpoolDir();
  chdir(proto);
 }
 
@@ -832,7 +867,7 @@ static void what_purge_compress_age(char *proto,char *host,char *file,int *purge
 /*++++++++++++++++++++++++++++++++++++++
   Compress the named file (if not already compressed).
 
-  long compress_file Returns the number of blocks of data compressed.
+  unsigned long compress_file Returns the number of blocks of data compressed.
 
   char *proto The protocol to use.
 
@@ -841,7 +876,7 @@ static void what_purge_compress_age(char *proto,char *host,char *file,int *purge
   char *file The name of the file in the current directory to compress.
   ++++++++++++++++++++++++++++++++++++++*/
 
-static long compress_file(char *proto,char *host,char *file)
+static unsigned long compress_file(char *proto,char *host,char *file)
 {
  int ifd,ofd;
  char *zfile;
@@ -865,7 +900,6 @@ static long compress_file(char *proto,char *host,char *file)
    }
 
  ifd=open(file,O_RDONLY|O_BINARY);
- init_buffer(ifd);
 
  if(ifd==-1)
    {
@@ -873,6 +907,8 @@ static long compress_file(char *proto,char *host,char *file)
     return(0);
    }
 
+ init_io(ifd);
+
  ParseReply(ifd,&spool_head);
 
  if(!spool_head ||
@@ -882,7 +918,10 @@ static long compress_file(char *proto,char *host,char *file)
    {
     if(spool_head)
        FreeHeader(spool_head);
+
+    finish_io(ifd);
     close(ifd);
+
     utime(file,&ubuf);
     return(0);
    }
@@ -896,31 +935,38 @@ static long compress_file(char *proto,char *host,char *file)
  strcat(zfile,".z");
 
  ofd=open(zfile,O_WRONLY|O_BINARY|O_CREAT|O_TRUNC,buf.st_mode&07777);
- init_buffer(ofd);
 
  if(ofd==-1)
    {
     PrintMessage(Inform,"Cannot open file '%s/%s/%s' to compress to [%!s].",proto,host,zfile);
+
     FreeHeader(spool_head);
+    free(zfile);
+
+    finish_io(ifd);
     close(ifd);
+
     utime(file,&ubuf);
     return(0);
    }
 
+ init_io(ofd);
+
  head=HeaderString(spool_head);
  FreeHeader(spool_head);
 
  write_string(ofd,head);
  free(head);
 
- init_zlib_buffer(ofd,-2);
+ configure_io_write(ofd,-1,2,0);
 
  while((n=read_data(ifd,buffer,READ_BUFFER_SIZE))>0)
     write_data(ofd,buffer,n);
 
- finish_zlib_buffer(ofd);
-
+ finish_io(ofd);
  close(ofd);
+
+ finish_io(ifd);
  close(ifd);
 
  if(rename(zfile,file))
@@ -936,6 +982,9 @@ static long compress_file(char *proto,char *host,char *file)
 
  free(zfile);
 
- return(Bytes_to_Blocks(orig_size-new_size));
+ if(new_size<orig_size)
+    return(Bytes_to_Blocks(orig_size-new_size));
+ else
+    return(0);                  /* pathological file, larger when compressed. */
 }
 #endif
